package server.command;

import java.util.Collection;

import common.collect.Lists;
import common.util.BlockPos;
import common.util.ExtMath;
import common.util.Position;

public class IntParser extends DefaultingParser {
	public static enum CoordType {
		X, Y, Z;
	}
	
	private final CoordType defType;
	private final Integer min;
	private final Integer max;
	private final boolean hex;
	
	public IntParser(String name, boolean hex, Integer def, Integer min, Integer max, Object ... completions) {
		super(name, def, completions);
		this.defType = null;
		this.min = min;
		this.max = max;
		this.hex = hex;
	}
	
	public IntParser(String name, CoordType type, Integer min, Integer max) {
		super(name, null);
		this.defType = type;
		this.min = min;
		this.max = max;
		this.hex = false;
	}
	
	public Integer parse(CommandEnvironment env, String input) {
		Integer pre = this.defType != null && input.startsWith("~") ? this.getDefault(env) : null;
		input = pre != null ? input.substring(1) : input;
		int value;
		if(pre != null && input.isEmpty()) {
			value = pre;
		}
		else {
			try {
				value = Integer.parseInt(input, this.hex ? 16 : 10);
			}
			catch(NumberFormatException e) {
				throw new RunException("Ungültige " + (this.hex ? "Hexadezimalzahl" : "Ganzzahl") + " '%s'", input);
			}
			if(pre != null)
				value = pre + value;
		}
		if(this.min != null && value < this.min)
			if(this.max != null)
				throw new RunException("Die Zahl muss im Bereich %d .. %d liegen, habe %d", this.min, this.max, value);
			else
				throw new RunException("Die Zahl muss mindestens %d betragen, habe %d", this.min, value);
		if(this.max != null && value > this.max)
			if(this.min != null)
				throw new RunException("Die Zahl muss im Bereich %d .. %d liegen, habe %d", this.min, this.max, value);
			else
				throw new RunException("Die Zahl darf höchstens %d betragen, habe %d", this.max, value);
		return value;
	}
	
	public Integer getDefault(CommandEnvironment env) {
		Position pos = this.defType == null ? null : env.getExecutor().getExecPos();
		if(this.defType != null)
			switch(this.defType) {
				case X:
					return pos == null ? null : ExtMath.floord(pos.x());
				case Y:
					return pos == null ? null : ExtMath.floord(pos.y());
				case Z:
					return pos == null ? null : ExtMath.floord(pos.z());
			}
		return (Integer)super.getDefault(env);
	}

	public Class<?> getTypeClass(boolean required) {
		return this.hasDefault() || required ? int.class : Integer.class;
	}
	
	public Collection<String> getCompletions(CommandEnvironment env) {
		BlockPos pos = this.defType == null ? null : env.getExecutor().getPointedPosition();
		if(this.defType != null)
			switch(this.defType) {
				case X:
					return pos == null ? null : Lists.newArrayList("" + pos.getX());
				case Y:
					return pos == null ? null : Lists.newArrayList("" + pos.getY());
				case Z:
					return pos == null ? null : Lists.newArrayList("" + pos.getZ());
			}
		return super.getCompletions(env);
	}
}
