package server.network;

import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.util.Arrays;

import javax.crypto.SecretKey;

import common.color.TextColor;
import common.init.Config;
import common.log.Log;
import common.net.util.concurrent.Future;
import common.net.util.concurrent.GenericFutureListener;
import common.network.ILoginHandler;
import common.network.IPlayer;
import common.network.NetConnection;
import common.network.NetHandler;
import common.packet.LPacketChallenge;
import common.packet.LPacketPassword;
import common.packet.LPacketPubkey;
import common.packet.LPacketResponse;
import common.packet.LPacketStartEncrypt;
import common.packet.RPacketChallenge;
import common.packet.RPacketDisconnect;
import common.packet.RPacketRequestEncrypt;
import common.packet.RPacketResponse;
import common.packet.RPacketServerConfig;
import server.Server;

public class LoginHandler extends NetHandler implements ILoginHandler
{
    private static enum LoginState {
        INIT, ENCRYPT, PROOF, PASSWORD, CHALLENGE, AUTHENTICATED, ACCEPTED;
    }
    
    private static final SecureRandom TOKEN_RNG = new SecureRandom();
    
    private final Server server;
    public final NetConnection netManager;
    
    private LoginState state = LoginState.INIT;
    private int timer;
    private String loginUser;
    private String loginPass;
    private byte[] loginToken;
    private PublicKey loginKey;

    public LoginHandler(Server server, NetConnection netManager)
    {
        this.netManager = netManager;
        this.server = server;
    }

    public void closeConnection(String reason)
    {
        try
        {
            Log.NETWORK.info("Trenne " + this.getConnectionInfo());
            this.netManager.sendPacket(new RPacketDisconnect(reason));
            this.netManager.closeChannel(reason);
        }
        catch (Exception exception)
        {
            Log.NETWORK.error(exception, "Fehler beim Trennen des Spielers beim Login");
        }
    }
    
    public void onDisconnect(String reason)
    {
        Log.NETWORK.info(this.getConnectionInfo() + " wurde beim Login getrennt: " + TextColor.stripCodes(reason));
    }

    public String getConnectionInfo()
    {
        return this.loginUser != null ? (this.loginUser + " (" + this.netManager.getCutAddress()
        		+ ")") : this.netManager.getCutAddress();
    }
    
    public void update()
    {
        if(this.state == LoginState.AUTHENTICATED)
        	this.tryAcceptPlayer();
//        else if (this.currentLoginState == LoginState.DELAY_ACCEPT)
//        {
//            if (this.server.getPlayer(this.loginUser) == null)
//            {
//                this.currentLoginState = LoginState.ACCEPTED;
//                this.server.addPlayer(this.networkManager, this.loginUser);
//            }
//        }
        if(this.timer++ == 600)
            this.closeConnection("Anmeldung dauerte zu lange");
    }

    private void tryAcceptPlayer()
    {
		if(this.server.getPlayer(this.loginUser) != null) {
			this.closeConnection("Nutzername '" + this.loginUser + "' ist bereits vergeben.");
			return;
		}
//            if (this.server.getPlayer(this.loginUser) != null)
//            {
//            	this.closeConnection("Nutzername '" + this.loginUser + "' ist bereits vergeben.");
//            	return;
////                this.currentLoginState = LoginState.DELAY_ACCEPT;
////                player.netHandler.kick("Du hast dich von einen anderen Ort verbunden");
//            }
//            this.networkManager.sendPacket(new RPacketLoginSuccess());
        String kick = this.server.addPlayer(this.netManager, this.loginUser, this.loginPass, this.loginKey);
        if(kick != null)
        	this.closeConnection(kick);
        else
            this.state = LoginState.ACCEPTED;
    }

    public void sendLoginPacket() {
		if(this.state != LoginState.INIT)
			throw new IllegalStateException("Unerwartetes Handshake-Paket");
		if(Config.encrypt) {
			this.state = LoginState.ENCRYPT;
			this.loginToken = new byte[4];
			TOKEN_RNG.nextBytes(this.loginToken);
	    	this.netManager.sendPacket(new RPacketRequestEncrypt(this.server.getPublicKey(), this.loginToken));
		}
		else {
			this.state = LoginState.PASSWORD;
			this.netManager.sendPacket(new RPacketServerConfig(Config.accessRequired, Config.authenticate, Config.authenticate && Config.passwordAuth, false));
		}
    }
    
	public void processEncryption(LPacketStartEncrypt packet) {
		if(this.state != LoginState.ENCRYPT)
			throw new IllegalStateException("Unerwartetes Verschlüsselungs-Paket");
        PrivateKey pkey = this.server.getPrivateKey();
        if(!Arrays.equals(this.loginToken, packet.getToken(pkey)))
            throw new IllegalStateException("Fehlerhaftes Token");
        SecretKey key = packet.getKey(pkey);
        this.netManager.startEncryption(key);
		this.state = LoginState.PROOF;
	}

	public void processChallenge(LPacketChallenge packet) {
		if(this.state != LoginState.PROOF)
			throw new IllegalStateException("Unerwartetes Anforderungs-Paket");
		this.state = LoginState.PASSWORD;
		this.netManager.sendPacket(new RPacketResponse(packet.getToken(this.server.getPrivateKey())), new GenericFutureListener < Future <? super Void >> () {
            public void operationComplete(Future <? super Void > u) throws Exception {
                LoginHandler.this.netManager.sendPacket(new RPacketServerConfig(Config.accessRequired, Config.authenticate, Config.authenticate && Config.passwordAuth, 
	    			Config.authenticate && Config.pubkeyAuth));
            }
        });
	}
	
	private boolean checkAccess(String access) {
		if(Config.accessRequired) {
	    	if(Config.password.length() < 8) {
	    		this.closeConnection("Es ist kein Zugangspasswort für diesen Server konfiguriert");
	    		return false;
			}
	    	if(!Config.password.equals(access)) {
	    		this.closeConnection("Falsches Zugangspasswort");
	    		return false;
			}
		}
		return true;
	}
    
    public void processPassword(LPacketPassword packet) {
		if(this.state != LoginState.PASSWORD)
			throw new IllegalStateException("Unerwartetes Passwort-Paket");
		this.loginUser = packet.getUser();
        if(this.loginUser.isEmpty() || !IPlayer.isValidUser(this.loginUser))
        	throw new IllegalStateException("Ungültiger Nutzername!");
		if(!Config.passwordAuth && Config.authenticate) {
			this.closeConnection("Dieser Server " + (Config.pubkeyAuth && Config.encrypt ? "benötigt einen öffentlichen Schlüssel zur Authentifizierung" : "hat keine Authentifizierungsmethode konfiguriert"));
			return;
		}
		if(!this.checkAccess(packet.getAccess()))
			return;
		if(Config.authenticate)
			this.loginPass = packet.getPassword();
		this.state = LoginState.AUTHENTICATED;
	}

	public void processPubkey(LPacketPubkey packet) {
		if(this.state != LoginState.PASSWORD)
			throw new IllegalStateException("Unerwartetes Pubkey-Paket");
		this.loginUser = packet.getUser();
        if(this.loginUser.isEmpty() || !IPlayer.isValidUser(this.loginUser))
        	throw new IllegalStateException("Ungültiger Nutzername!");
		if((!Config.pubkeyAuth || !Config.encrypt) && Config.authenticate) {
			this.closeConnection("Dieser Server " + (Config.passwordAuth ? "benötigt ein Passwort zur Authentifizierung" : "hat keine Authentifizierungsmethode konfiguriert"));
			return;
		}
		if(!this.checkAccess(packet.getAccess()))
			return;
    	if(Config.authenticate) {
			this.loginKey = packet.getKey();
			this.loginToken = new byte[32];
			TOKEN_RNG.nextBytes(this.loginToken);
			this.netManager.sendPacket(new RPacketChallenge(this.loginKey, this.loginToken));
			this.state = LoginState.CHALLENGE;
    	}
    	else {
    		this.state = LoginState.AUTHENTICATED;
    	}
	}

	public void processResponse(LPacketResponse packet) {
		if(this.state != LoginState.CHALLENGE)
			throw new IllegalStateException("Unerwartetes Beweis-Paket");
		if(!Arrays.equals(this.loginToken, packet.getToken()))
            throw new IllegalStateException("Fehlerhaftes Beweis-Token");
		this.state = LoginState.AUTHENTICATED;
	}
}
