package server.world;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.zip.DeflaterOutputStream;
import java.util.zip.InflaterInputStream;

import common.block.Block;
import common.collect.Lists;
import common.collect.Maps;
import common.entity.Entity;
import common.init.BlockRegistry;
import common.init.EntityRegistry;
import common.log.Log;
import common.tags.TagObject;
import common.tileentity.TileEntity;
import common.util.BlockPos;
import common.util.NextTickListEntry;
import common.util.NibbleArray;
import common.util.Util;
import common.world.BlockArray;

public class Region {
	private static class ChunkBuffer extends ByteArrayOutputStream {
		public ChunkBuffer() {
			super(8096);
		}

		public byte[] getData() {
			return this.buf;
		}
	}
	
	static {
		Thread thread = new Thread(new Runnable() {
			public void run() {
				while(!killed) {
					processQueue();
				}
			}
		}, "File IO Thread");
		thread.setPriority(1);
		thread.start();
	}
	
	private static final Map<String, Region> CACHE = Maps.<String, Region>newHashMap();
	private static final List<WorldServer> QUEUE = Collections.<WorldServer>synchronizedList(Lists.<WorldServer>newArrayList());
	
	private static volatile long queued;
	private static volatile long saved;
	private static volatile boolean waiting;
	private static volatile boolean killed;

	private final int[] timestamps = new int[64];
	private final int[] positions = new int[64];
	private final int[] sizes = new int[64];
	private final File regFile;
	private final File folder;
	private final int xPos;
	private final int zPos;

	private RandomAccessFile file;
	private long lastModified;
	private int offset;
	private boolean modified;

	public Region(File dir, int x, int z) {
		File sdir = new File(dir, Util.getRegionFolder(x << 3, z << 3));
		if(!sdir.exists())
			sdir.mkdirs();
		this.regFile = new File(sdir, Util.getRegionName(x << 3, z << 3));
		this.folder = dir;
		this.xPos = x;
		this.zPos = z;
		try {
			if(this.regFile.exists())
				this.lastModified = this.regFile.lastModified();
			this.file = new RandomAccessFile(this.regFile, "rw");
			this.file.seek(0L);
			if(this.file.length() < 512L) {
				for(int i = 0; i < 64; i++) {
					this.file.writeLong(0L);
				}
			}
			else {
				for(int i = 0; i < 64; i++) {
					int end = this.positions[i] = this.file.readUnsignedShort();
					end += ((this.sizes[i] = this.file.readUnsignedShort()) + 255) >> 8;
					this.timestamps[i] = this.file.readInt();
					if(end > this.offset)
						this.offset = end;
				}
			}
		}
		catch(IOException e) {
			e.printStackTrace();
		}
	}
	
	private void reorganize(boolean reopen) throws IOException {
		byte[] buffer = new byte[8192];
		File tmp = new File(this.folder, this.regFile.getName() + ".tmp");
		RandomAccessFile file = new RandomAccessFile(tmp, "rw");
		int offset = 0;
		for(int i = 0; i < 64; i++) {
			file.writeShort(this.sizes[i] == 0 ? 0 : offset);
			file.writeShort(this.sizes[i]);
			file.writeInt(this.timestamps[i]);
			offset += (this.sizes[i] + 255) >> 8;
		}
		this.offset = offset;
		offset = 0;
		for(int i = 0; i < 64; i++) {
			int size = this.sizes[i];
			if(size == 0) {
				this.positions[i] = 0;
				continue;
			}
			this.file.seek(512L + (long)(this.positions[i] << 8));
			this.file.read(buffer, 0, size);
			file.write(buffer, 0, size);
			this.positions[i] = offset;
			offset += (size + 255) >> 8;
			file.seek(512L + (long)(offset << 8));
		}
		file.close();
		this.file.close();
		tmp.renameTo(this.regFile);
		this.file = reopen ? new RandomAccessFile(this.regFile, "rw") : null;
		this.modified = false;
	}

	private synchronized byte[] read(int x, int z) {
		if(this.timestamps[x + z * 8] == 0)
			return null;
		FileInputStream in = null;
		try {
//			this.file.seek((long)(512 + (x + z * 8) * 8192));
			int size = this.sizes[x + z * 8]; // this.file.readShort();
			if(size > 8192 /* - 2 */ || size < 0) {
				Log.IO.warn("Chunk-Region-Datei " + this.regFile + " hat eine ungültige Größe bei " + x + ", " + z + ", überspringe");
				return null;
			}
			byte[] data;
			if(size == 0) {
				File expand = getExpansionFile(this.folder, this.xPos * 8 + x, this.zPos * 8 + z);
				if(!expand.exists()) {
					Log.IO.warn("Chunk-Erweiterungs-Datei " + expand + " ist nicht vorhanden oder nicht lesbar, überspringe");
					return null;
				}
				in = new FileInputStream(expand);
				data = new byte[(int)expand.length()];
				int remain = data.length;
				while(remain > 0) {
					int n = Math.min(remain, 8192);
					in.read(data, (data.length - remain), n);
					remain -= n;
				}
				try {
					if(in != null) {
						in.close();
					}
				}
				catch(IOException e) {
				}
				in = null;
			}
			else {
				int pos = this.positions[x + z * 8] << 8;
				if(pos + size > this.file.length() - 512L || pos < 0) {
					Log.IO.warn("Chunk-Region-Datei " + this.regFile + " hat eine ungültige Position bei " + x + ", " + z + ", überspringe");
					return null;
				}
				this.file.seek(512L + (long)pos);
				data = new byte[size]; //  - 2];
				this.file.read(data);
			}
			return data;
		}
		catch(IOException e) {
			try {
				if(in != null) {
					in.close();
				}
			}
			catch(IOException e1) {
			}
			Log.IO.error(e, "Fehler beim lesen von Chunk-Region-Datei " + this.regFile + " bei " + x + ", " + z + ", überspringe");
			return null;
		}
	}

	private synchronized void write(int x, int z, byte[] data, int size) {
		try {
			if(((int)this.file.length() - 512) >= 1048576)
				this.reorganize(true);
			this.modified = true;
//			if(size >= 8192)
//				Log.DATA.info("REG " + x + ", " + z + ": " + size);
			int pos = 0;
			if(size /* + 2 */ > 8192) {
				FileOutputStream out = null;
				try {
					out = new FileOutputStream(getExpansionFile(this.folder, this.xPos * 8 + x, this.zPos * 8 + z));
					int remain = size;
					while(remain > 0) {
						int n = Math.min(remain, 8192);
						out.write(data, (size - remain), n);
						remain -= n;
					}
				}
				catch(IOException e) {
					e.printStackTrace();
				}
				try {
					if(out != null) {
						out.close();
					}
				}
				catch(IOException e) {
				}
				data = new byte[0];
				size = 0;
			}
			else {
				pos = this.offset;
				this.offset += (size + 255) >> 8;
				this.file.seek(512L + (long)(pos << 8)); // (512 + (x + z * 8) * 8192));
	//			this.file.writeShort(size);
				this.file.write(data, 0, size);
			}
			int time = (int)(System.currentTimeMillis() / 10000L);
			this.timestamps[x + z * 8] = time;
			this.positions[x + z * 8] = pos;
			this.sizes[x + z * 8] = size;
			this.file.seek((long)((x + z * 8) * 8));
			this.file.writeShort(pos);
			this.file.writeShort(size);
			this.file.writeInt(time);
		}
		catch(IOException e) {
			e.printStackTrace();
		}
	}

	public synchronized void close(boolean reorg) throws IOException { // FIX ??
		if(this.file != null) {
			if(reorg && this.modified)
				this.reorganize(false);
			else
				this.file.close();
			this.file = null;
		}
	}

	public long getTimestamp(int x, int z) {
		return (long)this.timestamps[x + z * 8] * 10000L;
	}

	public void writeTag(int x, int z, TagObject tag) throws IOException {
		ChunkBuffer buf = new ChunkBuffer();
		DataOutputStream out = new DataOutputStream(new DeflaterOutputStream(buf));
		TagObject.write(tag, out);
		out.close();
		this.write(x, z, buf.getData(), buf.size());
	}
	
	public File getFile() {
		return this.regFile;
	}

	private static synchronized Region getRegionFile(File dir, int x, int z) {
		Region reg = CACHE.get(dir + "." + x + "." + z);
		if(reg != null)
			return reg;
		if(CACHE.size() >= 256)
			clearCache(false);
		Region nreg = new Region(dir, x, z);
		CACHE.put(dir + "." + x + "." + z, nreg);
		return nreg;
	}

	private static File getExpansionFile(File dir, int x, int z) {
		File sdir = new File(dir, Util.getRegionFolder(x, z));
		if(!sdir.exists())
			sdir.mkdirs();
		return new File(sdir, String.format("c.%c%X%c%X.chk", x < 0 ? 'n' : 'p', (x < 0) ? -x : x, z < 0 ? 'n' : 'p', (z < 0) ? -z : z));
	}

	private static synchronized void clearCache(boolean all) {
		if(all) {
			for(Region reg : CACHE.values()) {
				try {
					reg.close(true);
				}
				catch(IOException e) {
					e.printStackTrace();
				}
			}
			CACHE.clear();
		}
		else {
			Iterator<Region> iter = CACHE.values().iterator();
			for(int z = 0; z < 8 && iter.hasNext(); z++) {
				try {
					iter.next().close(true);
				}
				catch(IOException e) {
					e.printStackTrace();
				}
				iter.remove();
			}
		}
	}

	public static void finishWrite() {
		waiting = true;
		try {
			while(queued != saved) {
				Thread.sleep(10L);
			}
		}
		catch(InterruptedException e) {
			e.printStackTrace();
		}
		waiting = false;
		clearCache(true);
	}

	public static /* synchronized */ TagObject readChunk(File dir, int x, int z) throws IOException {
//		return getRegionFile(dir, x >> 3, z >> 3).readTag(x & 7, z & 7);
		byte[] data = getRegionFile(dir, x >> 3, z >> 3).read(x & 7, z & 7);
		if(data == null)
			return null;
		return TagObject.read(new DataInputStream(new BufferedInputStream(new InflaterInputStream(new ByteArrayInputStream(data)))));
	}

	public static /* synchronized */ void writeChunk(File dir, int x, int z, TagObject tag) throws IOException {
		ChunkBuffer buf = new ChunkBuffer();
		DataOutputStream out = new DataOutputStream(new DeflaterOutputStream(buf));
		TagObject.write(tag, out);
		out.close();
		getRegionFile(dir, x >> 3, z >> 3).write(x & 7, z & 7, buf.getData(), buf.size());
//		getRegionFile(dir, x >> 3, z >> 3).writeTag(x & 7, z & 7, tag);
	}
	
	public static ChunkServer readChunk(WorldServer world, int x, int z, TagObject tag) {
//		if(!tag.hasTag("Level")) {
//			Log.error("Chunk-Datei bei " + x + "," + z + " hat keine Level-Daten, überspringe");
//			return null;
//		}
//		tag = tag.getCompoundTag("Level");
		if(!tag.hasList("Sections")) {
			Log.IO.warn("Chunk-Datei bei " + x + "," + z + " hat keine Block-Daten, überspringe");
			return null;
		}
		ChunkServer chunk = new ChunkServer(world, x, z);
		chunk.setHeights(tag.getIntArray("HeightMap"));
		chunk.setTerrainPopulated(tag.getBool("TerrainPopulated"));
		chunk.setLightPopulated(tag.getBool("LightPopulated"));
		chunk.setInhabited(tag.getLong("InhabitedTime"));
		List<TagObject> sects = tag.getList("Sections");
		BlockArray[] sections = new BlockArray[sects.size()];
		boolean light = !world.dimension.hasNoLight();

		for(int n = 0; n < sects.size(); ++n) {
			TagObject sect = sects.get(n);
			int y = sect.getInt("Y");
			BlockArray storage = new BlockArray(y << 4, light, null);
			byte[] blocks = sect.getByteArray("Blocks");
			NibbleArray data = new NibbleArray(sect.getByteArray("Data"));
			NibbleArray adddata = sect.hasByteArray("Add") ? new NibbleArray(sect.getByteArray("Add")) : null;
			char[] seg = new char[blocks.length];

			for(int c = 0; c < seg.length; ++c) {
				int cx = c & 15;
				int cy = c >> 8 & 15;
				int cz = c >> 4 & 15;
				int ca = adddata != null ? adddata.get(cx, cy, cz) : 0;
				seg[c] = (char)(ca << 12 | (blocks[c] & 255) << 4 | data.get(cx, cy, cz));
			}

			storage.setData(seg);
			storage.setBlocklight(new NibbleArray(sect.getByteArray("BlockLight")));

			if(light) {
				storage.setSkylight(new NibbleArray(sect.getByteArray("SkyLight")));
			}

			storage.update();
			sections[n] = storage;
		}

		chunk.setStorage(sections);

		if(tag.hasByteArray("Biomes")) {
			chunk.setBiomes(tag.getByteArray("Biomes"));
		}

		List<TagObject> entities = tag.getList("Entities");

		if(entities != null) {
			for(int n = 0; n < entities.size(); ++n) {
				TagObject ent = entities.get(n);
				Entity entity = EntityRegistry.createFromTags(ent, world);
				chunk.setHasEntities(true);

				if(entity != null) {
					chunk.addEntity(entity);
					Entity rider = entity;

					for(TagObject ride = ent; ride.hasObject("Riding"); ride = ride.getObject("Riding")) {
						Entity pass = EntityRegistry.createFromTags(ride.getObject("Riding"), world);

						if(pass != null) {
							chunk.addEntity(pass);
							rider.mountEntity(pass);
						}

						rider = pass;
					}
				}
			}
		}

		List<TagObject> tiles = tag.getList("TileEntities");

		if(tiles != null) {
			for(int n = 0; n < tiles.size(); ++n) {
				TagObject tile = tiles.get(n);
				TileEntity tileentity = TileEntity.createAndLoadEntity(tile);

				if(tileentity != null) {
					chunk.addTileEntity(tileentity);
				}
			}
		}

		if(tag.hasList("TileTicks")) {
			List<TagObject> ticks = tag.getList("TileTicks");

			if(ticks != null) {
				int invalid = 0;
				for(int n = 0; n < ticks.size(); ++n) {
					TagObject tick = ticks.get(n);
					Block block;

					if(tick.hasString("i")) {
						block = BlockRegistry.getByIdFallback(tick.getString("i"));
					}
					else {
						block = BlockRegistry.getBlockById(tick.getInt("i"));
					}

					if(block != null) { // FIX
						world.scheduleBlockUpdate(new BlockPos(tick.getInt("x"), tick.getInt("y"), tick.getInt("z")), block,
								tick.getInt("t"), tick.getInt("p"));
					}
					else if(invalid++ < 10) {
						Log.IO.warn("Unbekannter Block-Tick in Chunk " + x + "," + z + ": " + 
						(tick.hasString("i") ? ("'" + tick.getString("i") + "'") : ("#" + tick.getInt("i"))));
					}
				}
				if(invalid > 10) {
					Log.IO.warn((invalid - 10) + " weitere ...");
				}
			}
		}

		return chunk;
	}

	public static TagObject writeChunk(WorldServer world, ChunkServer chunk) {
		TagObject tag = new TagObject();
//		tag.setShort("V", (short)Config.PROTOCOL);
		tag.setLong("LastUpdate", world.getTime());
		tag.setIntArray("HeightMap", chunk.getHeights());
		tag.setBool("TerrainPopulated", chunk.isTerrainPopulated());
		tag.setBool("LightPopulated", chunk.isLightPopulated());
		tag.setLong("InhabitedTime", chunk.getInhabited());
		Set<BlockArray> sections = chunk.getStorage();
		List<TagObject> sects = Lists.newArrayList();
		boolean light = !world.dimension.hasNoLight();

		for(BlockArray storage : sections) {
			if(storage != null) {
				TagObject sect = new TagObject();
				sect.setInt("Y", storage.getY() >> 4);
				byte[] blocks = new byte[storage.getData().length];
				NibbleArray data = new NibbleArray();
				NibbleArray adddata = null;

				for(int c = 0; c < storage.getData().length; ++c) {
					char cd = storage.getData()[c];
					int cx = c & 15;
					int cy = c >> 8 & 15;
					int cz = c >> 4 & 15;

					if(cd >> 12 != 0) {
						if(adddata == null) {
							adddata = new NibbleArray();
						}

						adddata.set(cx, cy, cz, cd >> 12);
					}

					blocks[c] = (byte)(cd >> 4 & 255);
					data.set(cx, cy, cz, cd & 15);
				}

				sect.setByteArray("Blocks", blocks);
				sect.setByteArray("Data", data.getData());

				if(adddata != null) {
					sect.setByteArray("Add", adddata.getData());
				}

				sect.setByteArray("BlockLight", storage.getBlocklight().getData());

				if(light) {
					sect.setByteArray("SkyLight", storage.getSkylight().getData());
				}
				else {
					sect.setByteArray("SkyLight", new byte[storage.getBlocklight().getData().length]);
				}

				sects.add(sect);
			}
		}

		tag.setList("Sections", sects);
		tag.setByteArray("Biomes", chunk.getBiomes());
		chunk.setHasEntities(false);
		List<TagObject> entities = Lists.newArrayList();

		for(int n = 0; n < chunk.getEntities().length; ++n) {
			for(Entity entity : chunk.getEntities()[n]) {
				TagObject ent = new TagObject();

				if(entity.writeOptional(ent)) {
					chunk.setHasEntities(true);
					entities.add(ent);
				}
			}
		}

		tag.setList("Entities", entities);
		List<TagObject> tiles = Lists.newArrayList();

		for(TileEntity tileentity : chunk.getTiles().values()) {
			TagObject tile = new TagObject();
			tileentity.writeTags(tile);
			tiles.add(tile);
		}

		tag.setList("TileEntities", tiles);
		List<NextTickListEntry> tics = world.getPendingBlockUpdates(chunk);

		if(tics != null) {
			long time = world.getTime();
			List<TagObject> ticks = Lists.newArrayList();

			for(NextTickListEntry tic : tics) {
				TagObject tick = new TagObject();
				String res = BlockRegistry.REGISTRY.getNameForObject(tic.getBlock());
				tick.setString("i", res == null ? "" : res.toString());
				tick.setInt("x", tic.position.getX());
				tick.setInt("y", tic.position.getY());
				tick.setInt("z", tic.position.getZ());
				tick.setInt("t", (int)(tic.scheduledTime - time));
				tick.setInt("p", tic.priority);
				ticks.add(tick);
			}

			tag.setList("TileTicks", ticks);
		}
		
		return tag;
	}

	private static void processQueue() {
		for(int i = 0; i < QUEUE.size(); ++i) {
			WorldServer loader = QUEUE.get(i);
			boolean flag = loader.writeNextIO();

			if(!flag) {
				QUEUE.remove(i--);
				++saved;
			}

			try {
				Thread.sleep(waiting ? 0L : 10L);
			}
			catch(InterruptedException e) {
				e.printStackTrace();
			}
		}

		if(QUEUE.isEmpty()) {
			try {
				Thread.sleep(25L);
			}
			catch(InterruptedException e) {
				e.printStackTrace();
			}
		}
	}

	public static void queueIO(WorldServer loader) {
		if(!QUEUE.contains(loader)) {
			++queued;
			QUEUE.add(loader);
		}
	}

	public static void killIO() {
		killed = true;
	}
}
