package server.worldgen.structure;

import java.util.Iterator;
import java.util.LinkedList;

import common.collect.Lists;
import common.rng.Random;
import common.tags.TagObject;
import java.util.List;
import common.util.ChunkPos;
import server.world.WorldServer;

public abstract class StructureStart
{
    protected LinkedList<StructureComponent> components = new LinkedList();
    protected StructureBoundingBox boundingBox;
    private int chunkPosX;
    private int chunkPosZ;

    public StructureStart()
    {
    }

    public StructureStart(int chunkX, int chunkZ)
    {
        this.chunkPosX = chunkX;
        this.chunkPosZ = chunkZ;
    }

    public StructureBoundingBox getBoundingBox()
    {
        return this.boundingBox;
    }

    public LinkedList<StructureComponent> getComponents()
    {
        return this.components;
    }

    /**
     * Keeps iterating Structure Pieces and spawning them until the checks tell it to stop
     */
    public void generateStructure(WorldServer worldIn, Random rand, StructureBoundingBox structurebb)
    {
        Iterator<StructureComponent> iterator = this.components.iterator();

        while (iterator.hasNext())
        {
            StructureComponent structurecomponent = (StructureComponent)iterator.next();

            if (structurecomponent.getBoundingBox().intersectsWith(structurebb) && !structurecomponent.addComponentParts(worldIn, rand, structurebb))
            {
                iterator.remove();
            }
        }
    }

    /**
     * Calculates total bounding box based on components' bounding boxes and saves it to boundingBox
     */
    protected void updateBoundingBox()
    {
        this.boundingBox = StructureBoundingBox.getNewBoundingBox();

        for (StructureComponent structurecomponent : this.components)
        {
            this.boundingBox.expandTo(structurecomponent.getBoundingBox());
        }
    }

    public TagObject writeComponents(int chunkX, int chunkZ)
    {
        TagObject tag = new TagObject();
        tag.setString("id", MapGenStructureIO.getStructureStartName(this));
        tag.setInt("ChunkX", chunkX);
        tag.setInt("ChunkZ", chunkZ);
        tag.setIntArray("BB", this.boundingBox.toIntArray());
        List<TagObject> list = Lists.newArrayList();

        for (StructureComponent comp : this.components)
        {
            list.add(comp.writeBase());
        }

        tag.setList("Children", list);
        this.writeTags(tag);
        return tag;
    }

    public void writeTags(TagObject tag)
    {
    }

    public void readComponents(WorldServer world, TagObject tag)
    {
        this.chunkPosX = tag.getInt("ChunkX");
        this.chunkPosZ = tag.getInt("ChunkZ");

        if (tag.hasIntArray("BB"))
        {
            this.boundingBox = new StructureBoundingBox(tag.getIntArray("BB"));
        }

        List<TagObject> list = tag.getList("Children");

        for (int i = 0; i < list.size(); ++i)
        {
            this.components.add(MapGenStructureIO.getStructureComponent(list.get(i), world));
        }

        this.readTags(tag);
    }

    public void readTags(TagObject tag)
    {
    }

    /**
     * offsets the structure Bounding Boxes up to a certain height, typically 63 - 10
     */
    protected void markAvailableHeight(WorldServer worldIn, Random rand, int p_75067_3_)
    {
        int i = worldIn.getSeaLevel() - p_75067_3_;
        int j = this.boundingBox.getYSize() + 1;

        if (j < i)
        {
            j += rand.zrange(i - j);
        }

        int k = j - this.boundingBox.maxY;
        this.boundingBox.offset(0, k, 0);

        for (StructureComponent structurecomponent : this.components)
        {
            structurecomponent.func_181138_a(0, k, 0);
        }
    }

    protected void setRandomHeight(WorldServer worldIn, Random rand, int p_75070_3_, int p_75070_4_)
    {
        int i = p_75070_4_ - p_75070_3_ + 1 - this.boundingBox.getYSize();
        int j = 1;

        if (i > 1)
        {
            j = p_75070_3_ + rand.zrange(i);
        }
        else
        {
            j = p_75070_3_;
        }

        int k = j - this.boundingBox.minY;
        this.boundingBox.offset(0, k, 0);

        for (StructureComponent structurecomponent : this.components)
        {
            structurecomponent.func_181138_a(0, k, 0);
        }
    }

    /**
     * currently only defined for Villages, returns true if Village has more than 2 non-road components
     */
    public boolean isSizeableStructure()
    {
        return true;
    }

    public boolean func_175788_a(ChunkPos pair)
    {
        return true;
    }

    public void func_175787_b(ChunkPos pair)
    {
    }

    public int getChunkPosX()
    {
        return this.chunkPosX;
    }

    public int getChunkPosZ()
    {
        return this.chunkPosZ;
    }
}
